#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from StringIO import StringIO
import tempfile
import unittest
import time
import sys
import os

import yaml

import usn


class SaveLoadDatabaseTest(unittest.TestCase):
    
    def setUp(self):
        self.filename = tempfile.mktemp()

    def tearDown(self):
        os.unlink(self.filename)

    def test_save_and_load_database(self):
        data = {"Hello": "world!"}
        usn.save_database(data, self.filename)
        self.assertTrue(os.path.isfile(self.filename))
        self.assertEquals(usn.load_database(self.filename), data)

    def test_save_and_load_database_with_filename(self):
        data = {"Hello": "world!"}
        usn.save_database(data, self.filename)
        self.assertTrue(os.path.isfile(self.filename))
        self.assertEquals(usn.load_database(self.filename), data)


class GetRestrictionsTest(unittest.TestCase):

    def setUp(self):
        self.original_filename = usn.CONFIG_FILENAME
        self.filename = usn.CONFIG_FILENAME = tempfile.mktemp()

    def tearDown(self):
        usn.CONFIG_FILENAME = self.original_filename
        if os.path.isfile(self.filename):
            os.unlink(self.filename)

    def test_unexistent(self):
        self.assertEquals(usn.get_restrictions(), {})

    def test_empty(self):
        file = open(self.filename, "w")
        file.write("[restrictions]\n")
        file.close()
        self.assertEquals(usn.get_restrictions(), {})

    def test_with_values(self):
        file = open(self.filename, "w")
        file.write("[restrictions]\n"
                   "option1 = value1 value2\n"
                   "option2 = value3 value4\n")
        file.close()
        self.assertEquals(usn.get_restrictions(),
                          {"option1": set(["value1", "value2"]),
                           "option2": set(["value3", "value4"])})


class USNMapTest(unittest.TestCase):

    def setUp(self):
        self.usn_map = usn.USNMap()

    def test_process_option_without_path(self):
        self.usn_map.process_option(None, "--foo", "bar", None)
        self.assertEquals(self.usn_map, {})

    def test_process_option_with_path(self):
        self.usn_map.process_option(None, "--foo", "bar", None, ("key",))
        self.assertEquals(self.usn_map, {"key": "bar"})

    def test_process_option_with_path_composed(self):
        self.usn_map.process_option(None, "--foo", "bar", None,
                                    ("key", "subkey"))
        self.assertEquals(self.usn_map, {"key": {"subkey": "bar"}})

    def test_process_option_with_two_paths(self):
        self.usn_map.process_option(None, "--foo", "bar", None,
                                    ("key", "subkey1"))
        self.usn_map.process_option(None, "--foo", "baz", None,
                                    ("key", "subkey2"))
        self.assertEquals(self.usn_map,
                          {"key": {"subkey1": "bar", "subkey2": "baz"}})

    def test_process_option_with_variable(self):
        self.usn_map.process_option(None, "--hey", "dude", None)
        self.usn_map.process_option(None, "--foo", "bar", None,
                                    ("key", usn.Var("hey"), "subkey"))
        self.assertEquals(self.usn_map, {"key": {"dude": {"subkey": "bar"}}})

    def test_process_option_append(self):
        self.usn_map.process_option(None, "--foo", "bar", None,
                                    ("key", "subkey"), append=True)
        self.assertEquals(self.usn_map, {"key": {"subkey": ["bar"]}})
        self.usn_map.process_option(None, "--foo", "baz", None,
                                    ("key", "subkey"), append=True)
        self.assertEquals(self.usn_map, {"key": {"subkey": ["bar", "baz"]}})

    def test_process_option_numeric(self):
        self.usn_map.process_option(None, "--foo", "123", None,
                                    ("key", "subkey"), numeric=True)
        self.assertEquals(self.usn_map, {"key": {"subkey": 123}})

    def test_allow_unrestricted(self):
        usn_map = usn.USNMap({"foo": set(["bar"])})
        usn_map.process_option(None, "--foo", "bar", None)

    def test_disallow_restricted(self):
        usn_map = usn.USNMap({"foo": set(["bar"])})
        self.assertRaises(usn.Error, usn_map.process_option,
                          None, "--foo", ("baz",), None)

    def test_process_option_from_stdin(self):
        sio = StringIO("STDIN")
        old_stdin = sys.stdin
        sys.stdin = sio
        try:
            self.usn_map.process_option(None, "--foo", "-", None, ("key",))
            self.assertEquals(self.usn_map, {"key": "STDIN"})
        finally:
            sys.stdin = old_stdin

    def test_out_of_order_error(self):
        try:
            self.usn_map.process_option(None, "--foo", "bar", None,
                                        (usn.Var("baz"),))
        except usn.Error, e:
            pass
        else:
            self.fail("Error not raised")
        self.assertEquals(str(e), "option --baz is needed before --foo")


class ParseOptionsTest(unittest.TestCase):

    def test_bad_args(self):
        self.assertRaises(usn.Error, usn.parse_options, ["123-4", "foo"])

    def test_lack_of_usn(self):
        self.assertRaises(usn.Error, usn.parse_options, ["--title", "foo"])

    def test_defaults(self):
        options = usn.parse_options([])
        self.assertEquals(options.usn, None)
        self.assertEquals(options.usn_map, {})

    def test_usn_map_is_dict(self):
        options = usn.parse_options(["123-4", "--title", "foo"])
        self.assertEquals(type(options.usn_map), dict)

    def test_usn(self):
        options = usn.parse_options(["123-4"])
        self.assertEquals(options.usn, "123-4")

    def test_template(self):
        options = usn.parse_options(["123-4", "--template", "TEMPL"])
        self.assertEquals(options.template, "TEMPL")

    def test_import(self):
        options = usn.parse_options(["--import"])
        self.assertTrue(options.import_)

    def test_export(self):
        options = usn.parse_options(["--export"])
        self.assertTrue(options.export)

    def test_title(self):
        options = usn.parse_options(["123-4", "--title", "TITLE"])
        self.assertEquals(options.usn_map, {"title": "TITLE"})

    def test_summary(self):
        options = usn.parse_options(["123-4", "--summary", "SUMMARY"])
        self.assertEquals(options.usn_map, {"summary": "SUMMARY"})

    def test_description(self):
        options = usn.parse_options(["123-4", "--description", "DESCR"])
        self.assertEquals(options.usn_map, {"description": "DESCR"})

    def test_action(self):
        options = usn.parse_options(["123-4", "--action", "ACTION"])
        self.assertEquals(options.usn_map, {"action": "ACTION"})

    def test_cve_one(self):
        options = usn.parse_options(["123-4", "--cve", "CVE"])
        self.assertEquals(options.usn_map, {"cves": ["CVE"]})

    def test_cve_two(self):
        options = usn.parse_options(["123-4", "--cve", "CVE1", "--cve", "CVE2"])
        self.assertEquals(options.usn_map, {"cves": ["CVE1", "CVE2"]})

    def test_timestamp(self):
        options = usn.parse_options(["123-4", "--timestamp", "123"])
        self.assertEquals(options.usn_map, {"timestamp": 123})

    def test_binary_version(self):
        options = usn.parse_options(["123-4", "--release", "REL",
                                     "--package", "PKG",
                                     "--binary-version", "VER"])
        self.assertEquals(options.usn_map,
                          {"releases":
                              {"REL":
                                  {"binaries": {"PKG": {"version": "VER"}}}}})

    def test_source_version(self):
        options = usn.parse_options(["123-4", "--release", "REL",
                                     "--package", "PKG",
                                     "--source-version", "VER"])
        self.assertEquals(options.usn_map,
                          {"releases":
                              {"REL":
                                  {"sources": {"PKG": {"version": "VER"}}}}})

    def test_url_size(self):
        options = usn.parse_options(["123-4", "--release", "REL",
                                     "--arch", "ARCH", "--url", "URL",
                                     "--url-size", "123"])
        self.assertEquals(options.usn_map,
                          {"releases":
                              {"REL":
                                  {"archs":
                                      {"ARCH":
                                          {"urls":
                                              {"URL": {"size": 123}}}}}}})

    def test_url_size(self):
        options = usn.parse_options(["123-4", "--release", "REL",
                                     "--arch", "ARCH", "--url", "URL",
                                     "--url-md5", "MD5"])
        self.assertEquals(options.usn_map,
                          {"releases":
                              {"REL":
                                  {"archs":
                                      {"ARCH":
                                          {"urls":
                                              {"URL": {"md5": "MD5"}}}}}}})

    def test_restriction(self):
        original_filename = usn.CONFIG_FILENAME
        filename = usn.CONFIG_FILENAME = tempfile.mktemp()
        file = open(filename, "w")
        file.write("[restrictions]\n"
                   "title = one two\n")
        file.close()
        try:
            options = usn.parse_options(["123-4", "--title", "one"])
            self.assertEquals(options.usn_map, {"title": "one"})
            options = usn.parse_options(["123-4", "--title", "two"])
            self.assertEquals(options.usn_map, {"title": "two"})
            self.assertRaises(usn.Error,
                              usn.parse_options, ["123-4", "--title", "three"])
            self.assertRaises(usn.Error,
                              usn.parse_options, ["123-4", "--title", "four"])
        finally:
            usn.CONFIG_FILENAME = original_filename
            os.unlink(filename)


class RenderTemplateTest(unittest.TestCase):

    def test_text_template(self):
        # We include some HTML to make sure that it's using the text markup.
        output = usn.render_template("Hello, $name!"
                                     '<p py:content="foo"></p>',
                                     {"name": "Bob"})
        self.assertEquals(output, 'Hello, Bob!<p py:content="foo"></p>')

    def test_html_template(self):
        # We include some text to make sure that it's using the HTML markup.
        output = usn.render_template("<p>{# hide #}\n"
                                     "#for name in names\n"
                                     "Hello, $name!\n"
                                     "#end\n"
                                     "</p>\n",
                                     {"name": "Bob", "names": ["Joe"]})
        self.assertEquals(output,
                          "<p>{# hide #}\n#for name in names\n"
                          "Hello, Bob!\n#end\n</p>")


class UpdateMapTest(unittest.TestCase):

    def test_do_nothing(self):
        map = {}
        changes = {}
        usn.update_map(map, changes)
        self.assertEquals(map, {})

    def test_add_key(self):
        map = {}
        changes = {"key": "value"}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": "value"})

    def test_add_key_and_subkey(self):
        map = {}
        changes = {"key": {"subkey": "value"}}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": {"subkey": "value"}})

    def test_add_subkey(self):
        map = {"key": {}}
        changes = {"key": {"subkey": "value"}}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": {"subkey": "value"}})

    def test_change_key(self):
        map = {"key": "value1"}
        changes = {"key": "value2"}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": "value2"})

    def test_change_subkey(self):
        map = {"key": {"subkey": "value1"}}
        changes = {"key": {"subkey": "value2"}}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": {"subkey": "value2"}})

    def test_extend_list(self):
        map = {"key": ["value1"]}
        changes = {"key": ["value2"]}
        usn.update_map(map, changes)
        self.assertEquals(map, {"key": ["value1", "value2"]})


class DecodeStringsTest(unittest.TestCase):

    def test_decode_strings(self):
        map = usn.decode_strings({"a": "\xc3\xa4", "b": 1, "c": {"d": "e"},
                              "f": ["g", "h", "i"]})
        self.assertEquals(map, {"a": u"ä", "b": 1, "c": {"d": "e"},
                                "f": ["g", "h", "i"]})
        self.assertEquals(type(map["a"]), unicode)
        self.assertEquals(type(map["c"]["d"]), unicode)
        self.assertEquals(type(map["f"][0]), unicode)
        self.assertEquals(type(map["f"][1]), unicode)
        self.assertEquals(type(map["f"][2]), unicode)


class MainTest(unittest.TestCase):

    def setUp(self):
        self.original_database_filename = usn.DATABASE_FILENAME
        self.original_config_filename = usn.CONFIG_FILENAME
        self.original_time = time.time
        self.database_filename = usn.DATABASE_FILENAME = tempfile.mktemp()
        self.config_filename = usn.CONFIG_FILENAME = tempfile.mktemp()
        self.time = 1187379812
        time.time = lambda: self.time

    def tearDown(self):
        usn.DATABASE_FILENAME = self.original_database_filename
        usn.CONFIG_FILENAME = self.original_config_filename
        time.time = self.original_time
        if os.path.isfile(self.database_filename):
            os.unlink(self.database_filename)
        if os.path.isfile(self.config_filename):
            os.unlink(self.config_filename)

    def test_create_usn(self):
        usn.main(["123-4", "--title", "one"])
        data = usn.load_database(self.database_filename)
        self.assertEquals(data,
                          {"123-4": {"id": "123-4",
                                     "timestamp": 1187379812,
                                     "title": "one",
                                     "releases": {}}})

    def test_update_usn(self):
        usn.save_database({"123-4": {"cves": ["CVE1"], "timestamp": 0}},
                          self.database_filename)
        usn.main(["123-4", "--title", "TITLE", "--cve", "CVE2"])
        data = usn.load_database(self.database_filename)
        self.assertEquals(data,
                          {"123-4": {"id": "123-4",
                                     "timestamp": 0,
                                     "title": "TITLE",
                                     "cves": ["CVE1", "CVE2"],
                                     "releases": {}}})

    def test_template(self):
        filename = tempfile.mktemp()
        try:
            file = open(filename, "w")
            file.write("Title: $title")
            file.close()
            usn.main(["123-4", "--title", "oné"])
            usn.main(["123-45", "--title", "two"])
            output = usn.main(["123-4", "--template", filename])
            self.assertEquals(output, "Title: oné")
            output = usn.main(["123-45", "--template", filename])
            self.assertEquals(output, "Title: two")
            self.assertRaises(usn.Error,
                              usn.main, ["123-456", "--template", filename])
        finally:
            os.unlink(filename)

    def test_timestamp_is_datetime(self):
        filename = tempfile.mktemp()
        try:
            file = open(filename, "w")
            file.write("$timestamp.__class__.__name__")
            file.close()
            usn.main(["123-4", "--timestamp", "0"])
            output = usn.main(["123-4", "--template", filename])
            self.assertEquals(output, "datetime")
        finally:
            os.unlink(filename)

    def test_import(self):
        usn.save_database({"123-4": {"cves": ["CVE1"]}},
                          self.database_filename)
        old_stdin = sys.stdin
        sys.stdin = StringIO("123-4: {cves: [CVE2]}")
        try:
            usn.main(["--import"])
        finally:
            sys.stdin = old_stdin
        data = usn.load_database(self.database_filename)
        self.assertEquals(data, {"123-4": {"cves": ["CVE2"]}})

    def test_export_without_usn(self):
        usn.save_database({"123-4": {"cves": ["CVE1"]},
                           "123-5": {"cves": ["CVE2"]}},
                          self.database_filename)
        self.assertEquals(usn.main(["--export"]),
                          "123-4:\n  cves: [CVE1]\n123-5:\n  cves: [CVE2]\n")

    def test_export(self):
        usn.save_database({"123-4": {"cves": ["CVE1"]}},
                          self.database_filename)
        self.assertEquals(usn.main(["--export"]),
                          "123-4:\n  cves: [CVE1]\n")


if __name__ == "__main__":
	unittest.main()
