#!/usr/bin/env python
from ConfigParser import ConfigParser
from optparse import OptionParser
from datetime import datetime, date
from string import Template
import cPickle
import pprint
import time
import sys
import os
import collections

import genshi.template
import yaml


DATABASE_FILENAME = "~/.usn.db"
CONFIG_FILENAME = "~/.usn.conf"


class Error(Exception):
    """Error class for application errors."""


class Var(str):
    """Marker for variables in paths."""


class USNMap(dict):

    def __init__(self, restrictions={}):
        self._restrictions = restrictions
        self._variables = {}

    def process_option(self, option, name, value, parser, path=None,
                       append=False, numeric=False):
        if value == "-":
            value = sys.stdin.read()
        name = name.lstrip("-")
        options = self._restrictions.get(name)
        if options and value not in options:
            raise Error("option --%s doesn't accept '%s' "
                        "according to your restrictions" % (name, value))
        self._variables[name] = value
        if numeric:
            value = int(value)
        if path is not None:
            path = list(path)
            for i in range(len(path)):
                elem = path[i]
                if type(elem) is Var:
                    if elem not in self._variables:
                        raise Error("option --%s is needed before --%s" %
                                    (elem, name))
                    path[i] = self._variables[elem]
            map = self
            for elem in path[:-1]:
                map = map.setdefault(elem, {})
            last_elem = path[-1]
            if append:
                if last_elem in map:
                    map[last_elem].append(value)
                else:
                    map[last_elem] = [value]
            else:
                map[last_elem] = value


def parse_options(argv):
    parser = OptionParser()
    parser.add_option("--show", action="store_true",
                      help="Show raw data for the given USN")
    parser.add_option("--show-cves", action="store_true",
                      help="Show CVEs for the given USN")
    parser.add_option("--show-sources", action="store_true",
                      help="Show source packages for the given USN")
    parser.add_option("--show-urls", action="store_true",
                      help="Show URLs for all updated files for the given USN")
    parser.add_option("--show-description", action="store_true",
                      help="Show description used in the given USN")
    parser.add_option("--show-title", action="store_true",
                      help="Show title used in the given USN")
    parser.add_option("--show-summary", action="store_true",
                      help="Show summary used in the given USN")
    parser.add_option("--show-action", action="store_true",
                      help="Show action used in the given USN")
    parser.add_option("--list", action="store_true",
                      help="Show list of known USNs")
    parser.add_option("--list-graph", action="store_true",
                      help="Show count of known USNs per month")
    parser.add_option("--export", action="store_true",
                      help="Show all USNs or selected one into YAML format")
    parser.add_option("--import", action="store_true", dest="import_",
                      help="Replace USNs with provided YAML data")
    parser.add_option("--template", action="store", metavar="FILENAME",
                      help="Render the USN, or all USNs if none specified, using the given template")
    parser.add_option("--output-dir", action="store", metavar="DIRECTORY",
                      help="Write template generated USN file(s) to the given directory (only valid with --template)")
    parser.add_option("--output-ext", action="store", metavar="EXTENSION",
                      help="Write template generated USN file(s) with the given filename extension (only valid with --output-dir)")
    parser.add_option("--db", action="store", metavar="FILENAME",
                      help="Use specified database file")

    usn_map = USNMap(get_restrictions())

    def register_option(name, help=None, metavar=None, **kwargs):
        parser.add_option(name, action="callback", type="string", help=help,
                          metavar=metavar, callback=usn_map.process_option,
                          callback_kwargs=kwargs)

    register_option("--title", path=("title",),
                    help="The shortest description for the USN")
    register_option("--summary", path=("summary",),
                    help="Normal description for the USN")
    register_option("--description", metavar="TEXT", path=("description",),
                    help="Textual description for the USN")
    register_option("--action", metavar="TEXT", path=("action",),
                    help="Explanation of what should be done")
    register_option("--issue-summary", metavar="TEXT", path=("isummary",),
                    help="High level explanation of issues")
    register_option("--timestamp", path=("timestamp",), numeric=True,
                    help="Timestamp when the USN was released "
                         "(defaults to now)")
    register_option("--cve", path=("cves",), append=True,
                    help="Link USN with the given CVE")
    register_option("--release", metavar="NAME",
                    help="Set Ubuntu release for options requiring it.")
    register_option("--arch",
                    help="Set architecture for options requiring it.")
    register_option("--package", metavar="NAME",
                    help="Set package name for options requiring it.")
    register_option("--binary-version", metavar="VER",
                    path=("releases", Var("release"), "binaries",
                          Var("package"), "version"),
                    help="Set binary package version with fix on the "
                         "given release")
    register_option("--source-version", metavar="VER",
                    path=("releases", Var("release"), "sources",
                          Var("package"), "version"),
                    help="Set source package version with fix on the "
                         "given release")
    register_option("--source-description", metavar="DESC",
                    path=("releases", Var("release"), "sources",
                          Var("package"), "description"),
                    help="Set source package description with fix on the "
                         "given release")
    register_option("--url", metavar="URL",
                    help="Add an URL linked to the current "
                         "release/architecture")
    register_option("--url-size", metavar="URL", numeric=True,
                    path=("releases", Var("release"), "archs", Var("arch"),
                          "urls", Var("url"), "size"),
                    help="Set size of the file in the previously provided URL")
    register_option("--url-md5", metavar="URL",
                    path=("releases", Var("release"), "archs", Var("arch"),
                          "urls", Var("url"), "md5"),
                    help="Set MD5 of the file in the previoulsy provided URL")

    options, args = parser.parse_args(argv)

    if len(args) == 1:
        options.usn = args[0]
    elif args:
        raise Error("incorrect number of arguments provided")
    else:
        options.usn = None

    if usn_map and not options.usn:
        raise Error("no USN identifier provided")
    else:
        options.usn_map = usn_map.copy()

    if options.output_dir and not options.template:
        parser.error("a template must be specified with an output directory")
    elif options.output_ext and not options.output_dir:
        parser.error("an output directory must be specified with an output extension")

    return options


def load_database(filename):
    filename = os.path.expanduser(filename)
    if not os.path.isfile(filename):
        return {}
    with open(filename) as f:
        database = cPickle.load(f)
    return database


def save_database(database, filename):
    # Make sure we don't destroy the existing database when saving the
    # new one and we run out of disk space or hit some other error.
    orig = os.path.expanduser(filename)
    name = orig + ".saving"
    with open(name, "w") as f:
        cPickle.dump(database, f, -1)
    os.rename(name, orig)


def get_restrictions():
    filename = os.path.expanduser(CONFIG_FILENAME)
    if not os.path.isfile(filename):
        return {}
    parser = ConfigParser()
    parser.read(filename)
    restrictions = {}
    for option in parser.options("restrictions"):
        restrictions[option] = set(parser.get("restrictions", option).split())
    return restrictions


def get_codename_to_version_dict(usn_timestamp):
    codename_to_version = collections.OrderedDict([
                        ('artful',      '17.10'),
                        ('zesty',       '17.04'),
                        ('yakkety',     '16.10'),
                        ('xenial',      '16.04 LTS'),
                        ('wily',        '15.10'),
                        ('vivid',       '15.04'),
                        ('utopic',      '14.10'),
                        ('trusty',      '14.04 LTS'),
                        ('saucy',       '13.10'),
                        ('raring',      '13.04'),
                        ('quantal',     '12.10'),
                        ('precise',     '12.04 LTS'),
                        ('oneiric',     '11.10'),
                        ('natty',       '11.04'),
                        ('maverick',    '10.10'),
                        ('lucid',       '10.04 LTS'),
                        ('karmic',      '9.10'),
                        ('jaunty',      '9.04'),
                        ('intrepid',    '8.10'),
                        ('hardy',       '8.04 LTS'),
                        ('gutsy',       '7.10'),
                        ('feisty',      '7.04'),
                        ('edgy',        '6.10'),
                        ('dapper',      '6.06 LTS'),
                        ('breezy',      '5.10'),
                        ('hoary',       '5.04'),
                        ('warty',       '4.10'),
                      ])

    # Set ESM status by comparing against the day after the release went EoL.
    # For example, 12.04 LTS went EoL on 2017-04-28 so all USNs issued on or
    # after 2017-04-29 use 12.04 ESM as the Ubuntu version.
    if usn_timestamp >= datetime(2017, 4, 29):
        codename_to_version['precise'] = '12.04 ESM'

    return codename_to_version


def is_esm_version(version):
    return version.endswith(' ESM')


def get_package_descriptions(releases, codename_to_version):
    descriptions = collections.OrderedDict([])

    for codename in codename_to_version.keys():
        if codename in releases and 'sources' in releases[codename]:
            for name in sorted(releases[codename]['sources']):
                description = releases[codename]['sources'][name].get('description', '')

                if name not in descriptions or (not descriptions[name] and description):
                    descriptions[name] = description

    return descriptions


def build_source_links(releases, codename_to_version):
    source_links = []

    for codename, version in codename_to_version.iteritems():
        # Don't show links for ESM releases since the links won't work
        if is_esm_version(version):
            continue

        if codename in releases and 'sources' in releases[codename]:
            for name in sorted(releases[codename]['sources']):
                link = 'https://launchpad.net/ubuntu/+source/%s/%s' % (name, releases[codename]['sources'][name]['version'])
                source_links.append(link)

    return source_links


def guess_binary_links(binary, version, sources):
    '''Guess links to the source package based on binary package and version.

    Keyword Arguments:
    binary -- the name of the binary
    version -- the version of the binary
    sources -- a dictionary of source package names to source package versions

    When successful, two different links are returned. The first is a link to
    the source package on Launchpad. The second is a link to the versioned
    source package on Launchpad. (None, None) is returned if a high quality
    cannot be made.

    This function was inspired by sanely_linkify() that was added by the
    following commit:

    revno: 43
    committer: Jamie Strandboge <jamie@canonical.com>
    message:
      first pass at new USN format

    It included this comment:

      Since there is no strict mapping of binary packages to source packages
      in the USN database, we cannot link back to the source packages for
      each binary package without adding external knowledge to this template.
      As a work-around, we can guess at the source package in the case where
      there is only 1 source package in the USN (common case), or when each
      source package has a distinct version number from every other binary
      package version number (common for multi-source USNs).  In strange
      situations, we can just fall back to an unlinkified USN report of
      packages.
    '''
    match_first = False
    source_match = None
    version_match = None
    source_link = None
    version_link = None

    if not sources:
        # Old USNs may not have any sources listed. We can't make any sort of a
        # guess in this situation.
        return (None, None)
    if len(sources) == 1:
        # There's a many-to-one mapping of binaries to a source package. Use
        # the only possible source package for all binaries.
        match_first = True
    elif not version:
        # There are multiple combinations of possible binary to source package
        # mappings. We can't make an educated guess if we don't have a valid
        # binary package version so don't attempt to construct a link.
        return (None, None)

    for source in sources:
        source_version = sources[source].get('version')
        if match_first or version == source_version:
            source_match = source
            version_match = source_version
            break

    if source_match:
        source_link = 'https://launchpad.net/ubuntu/+source/' + source_match
        if version_match:
            # Be certain to use the source package version rather than the
            # binary package version here or the link will be broken for
            # certain packages
            version_link = source_link + '/' + version_match

    return (source_link, version_link)


def insert_guessed_binary_links_for_release(release):
    '''Guess and insert links to source packages for all binary packages.

    Keyword Arguments:
    release -- a dict of binaries and sources for a given release

    See guess_binary_links() for caveats.
    '''
    for binary, details in release['binaries'].iteritems():
        (source_link, version_link) = guess_binary_links(binary,
                                                         details.get('version'),
                                                         release.get('sources'))
        release['binaries'][binary]['source_link'] = source_link
        release['binaries'][binary]['version_link'] = version_link


def insert_empty_binary_links_for_release(release):
    '''Insert empty (None) links to source packages for all binary packages.

    Keyword Arguments:
    release -- a dict of binaries and sources for a given release

    This is useful for binaries of an ESM release since any guessed links would
    be dead links that do not work.
    '''
    for binary in release['binaries']:
        release['binaries'][binary]['source_link'] = None
        release['binaries'][binary]['version_link'] = None


def insert_binary_links(releases, codename_to_version):
    '''Insert links to source packages for all binary packages in all releases.

    Keyword Arguments:
    releases -- a dict of releases containing dicts of binaries and sources
    codename_to_version -- a mapping of Ubuntu release codenames to versions

    Updates the releases dict with per-binary source package links. Two
    different classes of links are inserted. The first is a link to the source
    package on Launchpad. The second is a link to the versioned source package
    on Launchpad. Some templates, such as webpage-markdown.txt, may find these
    links useful.
    '''
    for codename, version in codename_to_version.iteritems():
        if codename not in releases:
            continue
        elif 'binaries' not in releases[codename]:
            continue
        elif is_esm_version(version):
            insert_empty_binary_links_for_release(releases[codename])
        else:
            insert_guessed_binary_links_for_release(releases[codename])


def render_template(template, data):
    if template.lstrip().startswith("<") and template.rstrip().endswith(">"):
        Template = genshi.template.MarkupTemplate
        format = "xhtml"
    else:
        Template = genshi.template.NewTextTemplate
        format = "text"
    stream = Template(template).generate(**data)
    return stream.render(format)


def update_map(map, changes):
    for key, value in changes.iteritems():
        if key not in map:
            map[key] = value
        elif type(value) is list:
            map[key].extend(value)
        elif type(value) is dict:
            update_map(map[key], value)
        else:
            map[key] = value


def decode_strings(obj):
    obj_type = type(obj)
    if obj_type is str:
        return obj.decode("utf-8", "replace")
    elif obj_type is dict:
        for key, value in list(obj.iteritems()):
            obj[key] = decode_strings(value)
    elif obj_type is list:
        for i in range(len(obj)):
            obj[i] = decode_strings(obj[i])
    return obj


def usncmp(a, b):
    return cmp(int(a.split('-')[0]), int(b.split('-')[0]))


def generate_usn_from_template(usn_map, template):
    usn_timestamp = datetime.utcfromtimestamp(usn_map["timestamp"])
    codename_to_version = get_codename_to_version_dict(usn_timestamp)

    usn_map["timestamp"] = usn_timestamp
    usn_map["source_links"] = build_source_links(usn_map['releases'], codename_to_version)
    insert_binary_links(usn_map['releases'], codename_to_version)
    usn_map["package_descriptions"] = get_package_descriptions(usn_map['releases'], codename_to_version)
    usn_map["codename_to_version"] = codename_to_version

    decode_strings(usn_map)  # Genshi doesn't like non-ascii plain strings.


    # XXX when people.c.c moves to a more modern python, get rid of this
    # hack. An example of a troublemaking usn is 3583-1.
    if sys.version_info.major == 2 and sys.version_info.minor == 7 and sys.version_info.micro < 6:
        return render_template(open(template).read(), usn_map)
    else:
        return render_template(open(template).read(), usn_map).encode('utf-8')


def main(argv):
    options = parse_options(argv)

    # Load database
    dbpath = DATABASE_FILENAME
    if options.db:
        dbpath = options.db
    database = load_database(dbpath)

    if options.import_:
        data = yaml.load(sys.stdin.read())
        database.update(data)
    if options.usn_map:
        usn_map = database.setdefault(options.usn, {})
        update_map(usn_map, options.usn_map)
        usn_map["id"] = options.usn
        usn_map.setdefault("releases", {})
        usn_map.setdefault("timestamp", time.time())
    if options.usn_map or options.import_:
        save_database(database, dbpath)
    if options.template:
        if options.usn:
            if options.usn not in database:
                raise Error("USN %s not in the database" % options.usn)

            usns = [options.usn]
        else:
            usns = sorted(database.keys(), cmp=usncmp)

        for usn in usns:
            usn_text = generate_usn_from_template(database[usn], options.template)

            if options.output_dir:
                path = os.path.join(options.output_dir, usn)
                if options.output_ext:
                    path += ".%s" % (options.output_ext)

                with open(path, "w") as f:
                    f.write(usn_text)
            else:
                sys.stdout.write(usn_text)
        return
    if options.export:
        if options.usn:
            data = dict()
            for usn in options.usn.split(','):
                data[usn] = database[usn]
        else:
            data = database
        return yaml.dump(data)
    if options.show or options.show_cves or options.show_sources or options.show_urls or options.show_summary or options.show_title or options.show_action or options.show_description:
        data = database
        if options.usn:
            if options.usn not in database:
                raise Error("USN %s not in the database" % options.usn)
            else:
                data = database[options.usn]
        if options.show:
            pp = pprint.PrettyPrinter(indent=4)
            pp.pprint(data)
        if options.show_cves:
            for cve in data['cves']:
                if cve.startswith('CVE'):
                    print cve
        if options.show_sources:
            sources = set()
            for rel in data['releases']:
                for source in data['releases'][rel]['sources']:
                    sources.add(source)
            for source in sources:
                print source
        if options.show_urls:
            urls = []
            for rel in data['releases']:
                for arch in data['releases'][rel]['archs'].keys():
                    urls += [] + data['releases'][rel]['archs'][arch]['urls'].keys()
            for url in urls:
                print url
        if options.show_description:
            print data['description']
        if options.show_title:
            print data['title']
        if options.show_summary:
            print data['summary']
        if options.show_action:
            print data['action']
    if options.list:
        for usn in sorted(database.keys(), cmp=usncmp):
            print usn
    if options.list_graph:
        months = dict()
        for usn in database.keys():
            month = datetime.utcfromtimestamp(database[usn]['timestamp']).strftime('%Y-%m')
            months.setdefault(month, 0)
            months[month] += 1
        for month in sorted(months):
            print "%s\t%d" % (month, months[month])


if __name__ == "__main__":
    try:
        output = main(sys.argv[1:])
        if output:
            print output.encode('utf-8')
    except Error, e:
        sys.exit("error: %s" % str(e))
